

   ### DEFINE HELPER FUNCTIONS
   # Calculate strata, when adjmodel = "stratified"
   getStrata <- function(in_d,in_y,in_adjcovMat,adjust=TRUE) {
      if(adjust) {
        # Evaluate number of strata
        nStrataVector <- vector(length=dim(in_adjcovMat)[2])
        nStrata <- 1
        for(i in 1:dim(in_adjcovMat)[2]) {
           nStrataVector[i] <- length(unique(in_adjcovMat[,i]))
           nStrata <- nStrata * nStrataVector[i]
        }

        # Create permutation matrix, to characterize each strata
        permMat <- matrix(nrow=nStrata, ncol=dim(in_adjcovMat)[2])
        nStrataOrdered <- nStrataVector[rev(order(nStrataVector))]
        adjcovNamesOrdered <-(colnames(in_adjcovMat))[rev(order(nStrataVector))]

        adjcovOrdered <- matrix(nrow=dim(in_adjcovMat)[1], 
                                ncol=dim(in_adjcovMat)[2])
        if(dim(in_adjcovMat)[2] > 1) {
           for(i in 1:(dim(in_adjcovMat)[2]-1)) {
              adjcovOrdered[,i] <- in_adjcovMat[,
                    which(colnames(in_adjcovMat) == adjcovNamesOrdered[i])]
              currStrata <- sort(unique(adjcovOrdered[,i]))
              permMat[,i] <- rep(currStrata, nStrata/nStrataOrdered[i] )
           }
        }
        adjcovOrdered[,dim(in_adjcovMat)[2]] <- in_adjcovMat[,
                                  which(colnames(in_adjcovMat) == 
                                  adjcovNamesOrdered[dim(in_adjcovMat)[2]])]
        currStrata <- sort(unique(adjcovOrdered[,dim(in_adjcovMat)[2]]))
        permMat[,dim(in_adjcovMat)[2]] <- rep(currStrata, 
                      each=nStrata/nStrataOrdered[dim(in_adjcovMat)[2]] )


        # Get subsets/strata from dataset
        strataList <- vector("list",length=nStrata)
        noCaseStrata <- NULL
        for(s in 1:nStrata) {
           currStrata <- as.data.frame(cbind(in_d,in_y,adjcovOrdered))
           #commented out on 8/19
           #colnames(currStrata) <- c("d","y",colnames(adjcovMat)) 
	   
	   # changed by DEM on 11/4 from d to "d"
           colnames(currStrata) <- c("d","marker",adjcovNamesOrdered)
           for(k in 1:dim(in_adjcovMat)[2]) 
              currStrata <- currStrata[which(eval(parse(
                      text=paste("currStrata$",adjcovNamesOrdered[k],
                      sep="")))==permMat[s,k]),]

   #        for(k in 1:dim(adjcovMat)[2]) {
   #            currStrata <- currStrata[which(adjcovOrdered[,k]==permMat[s,k]),]
   ##           currStrata <- subset(currStrata, currStrata[,2+k]==permMat[s,k])
   ##             eval(parse(text=paste("currStrata$",adjcovOrdered[k],sep="")))
   ##             adjcovMat[,which(names(adjcov)==adjcovOrdered[i])]
   #        }
           # Only include strata with atleast one case
           cases <- NULL
	   # DEM 11/4 ... goes with above change.
       #    eval(parse(text=paste("cases <- subset(currStrata,",d,"==1)",sep="")))
           cases <- subset(currStrata,d==1)

           #cases <- subset(currStrata,d==1)
           if(dim(cases)[1] > 0)
              strataList[[s]] <- currStrata
           else {
              strataList[[s]] <- NULL
              noCaseStrata <- c(noCaseStrata, s)
           }
        }
        if(is.null(noCaseStrata)) {
           return(strataList)
        } else {
           return(strataList[-noCaseStrata])
        }
      } else {
        # build list with full dataset
         currStrata <- as.data.frame(cbind(in_d,in_y,adjcovOrdered))
	 # DEM 11/4
         colnames(currStrata) <- c("d","marker",adjcovNamesOrdered)
         strataList <- list(currStrata)
         return(strataList)
      }   

   }



   # If adjusting for covariates using stratification, check that each case-containing (relevant)
   # strata has at least the specified number of controls (minCtrl)
   # d is disease vector, adjcovMat is matrix of covariates
   isStrataCtrlLow <- function(d,adjcovMat,minCtrl) {
      if(is.null(adjcovMat))
         return(FALSE)

      # Evaluate number of strata
      nStrataVector <- vector(length=dim(adjcovMat)[2])
      nStrata <- 1
      for(i in 1:dim(adjcovMat)[2]) {
         nStrataVector[i] <- length(unique(adjcovMat[,i]))
         nStrata <- nStrata * nStrataVector[i]
      }

      # Create permutation matrix, to characterize each strata
      permMat <- matrix(nrow=nStrata, ncol=dim(adjcovMat)[2])
      nStrataOrdered <- nStrataVector[rev(order(nStrataVector))]
      adjcovNamesOrdered <- (colnames(adjcovMat))[rev(order(nStrataVector))]

      adjcovOrdered <- matrix(nrow=dim(adjcovMat)[1], ncol=dim(adjcovMat)[2])
      for(i in 1:dim(adjcovMat)[2]) {
         adjcovOrdered[,i] <- adjcovMat[,
	                      which(colnames(adjcovMat)==adjcovNamesOrdered[i])]
         currStrata <- sort(unique(adjcovOrdered[,i]))
         permMat[,i] <- rep(currStrata, nStrata/nStrataOrdered[i] )
      }

      # Get subsets/strata from dataset
      for(s in 1:nStrata) {
         currStrata <- as.data.frame(cbind(d,adjcovOrdered))
         colnames(currStrata) <- c("d",adjcovNamesOrdered)
         for(k in 1:dim(adjcovMat)[2]) 
            currStrata <- currStrata[which(eval(parse(text=paste("currStrata$",
               adjcovNamesOrdered[k],sep="")))==permMat[s,k]),]
         cases <- subset(currStrata,d==1)
         if(dim(cases)[1] > 0) {
            ctrls <- subset(currStrata,d==0)
            if(dim(ctrls)[1] < minCtrl)
               return(TRUE)
         }
      }
      return(FALSE)
   }

   
   # Calculate percentile values using empirical method
   # Called by pcval
   pcval_calc <- function(d, y, tiecorr) {
      n_db <- length(d[d==0])
 #    Y_db <- sort(y[1:n_db])
 #     y_j <- sort(y[(n_db+1):length(y)])
      Y_db <- y[1:n_db]
      y_j <-  y[(n_db+1):length(y)]

      # Calculate P(Y_db < y_j)
      y_j_neg <- -(y_j)
      Y_db_neg <- -(Y_db)
      F_l_neg <- ecdf(Y_db_neg)
      F_l <- rep(1, length(y_j)) - F_l_neg(y_j_neg)

      if(!tiecorr) {
         pcv_j <- F_l         # P(Y_db < y_j)
      } else {
         # Calculate P(Y_db <= y_j)
         F_le <- ecdf(Y_db) 
         pcv_j <- F_l + 0.5*(F_le(y_j) - F_l)   # P(Y_db < y_j) + 0.5*P(Y_db = y_j)
      }
      return(pcv_j)
   }




   # Calculate percentile values using parametric (normal distn) method
   # Called by npcval 
   npcval_calc <- function(d, y) {
      n_db <- length(d[d==0])
      Y_db <- y[1:n_db]
      y_j  <- y[(n_db+1):length(y)]

      mu <- mean(Y_db)
      sigma <- sd(Y_db)

      # Calculate CDF of y_j, P(Y_db <= y_j), under Normality assumption
      F_le <- pnorm((y_j-mu)/sigma)
      return(F_le)
   }   
  
  
  
   #  This function takes a vector of cluster IDs and returns a 2-item list:
   #    (1) the unique sorted list of cluster IDs
   #    (2) a list where each element is a vector of the location(s) 
   #        of each cluster in the original vector (the first element in
   #        the list corresponds to the lowest cluster number)
   makeClist <- function(cVec) {
     lOut <- list()
     oVec <- order(cVec)
     sVec <- sort(cVec)
     j = 1
     lOut[[j]] <- oVec[1]
     for(i in 2:length(cVec)) {
       if(sVec[i] == sVec[i-1]) {
         lOut[[j]] <- c(lOut[[j]],oVec[i])
       } else {
         j = j + 1
         lOut[[j]] <- oVec[i]
       }
     }
     return(list(sVec = unique(sVec),lOut=lOut))    
   }
 
   

   getBstrapSample <- function(in_data,noccsamp, cluster,
                               clusterUniqueMat, cList, strataCluster=NULL) {
      if(!noccsamp) {         # case-control
         if(is.null(cluster)) {
            #n_db <- dim(subset(dataset_ordered, d==0))[1]
            #n_db <- length(d_vector[d_vector==0])
     #       eval(parse(text=paste("n_db<-dim(subset(in_data, ",d,"==0))[1]", 
     #                             sep="")))
            n_db <- length(which(in_data[,1]==0))		      
            resample.ctrl <- sample(1:n_db, replace=TRUE)
            resample.case <- sample((n_db+1):dim(in_data)[1], replace=TRUE)
         }
         else {
            if(is.null(strataCluster)) {
               strataCUMat = clusterUniqueMat
            } else {
               strataCUMat = clusterUniqueMat[cList$sVec,]
            }
            cVecs       = cList$lOut	   
    #        n_db <- dim(subset(strataCUMat,d==0))[1]
            n_db <- length(which(strataCUMat[,1]==0))	    
            resample.ctrlCluster <- sample(1:n_db, replace=TRUE)
            resample.caseCluster <- sample((n_db+1):dim(strataCUMat)[1], 
                                           replace=TRUE)
            resample.ctrl <- unlist(sapply(resample.ctrlCluster,
                                    function(x) cVecs[[x]]))
            resample.case <- unlist(sapply(resample.caseCluster,
                                    function(x) cVecs[[x]]))
         }
         return(list(resample.ctrl, resample.case))
      }
      else {                  # cohort
         if(is.null(cluster))
            resample.cohort <- sample(1:dim(in_data)[1], replace=TRUE)
         else {
            if(is.null(strataCluster)) {
               strataCUMat = clusterUniqueMat
            } else {
               strataCUMat = clusterUniqueMat[cList$sVec,]
            }
            cVecs  = cList$lOut	   
            resample.cohortCluster <- sample(1:dim(strataCUMat)[1], 
                                             replace=TRUE)
            resample.cohort <- unlist(sapply(resample.cohortCluster,
                                    function(x) cVecs[[x]]))
         }
         return(resample.cohort)
      }
   }

  
  

